@@ -1,4 +1,4 @@
-use std::{env, fmt, iter::Peekable, mem};
+use std::{env, ffi::OsString, fmt, iter::Peekable, mem};
 
 use anyhow::{bail, format_err, Error, Result};
 
@@ -69,23 +69,31 @@ pub(crate) struct Args<'a> {
     pub(crate) no_default_features: bool,
     /// -v, --verbose, -vv
     pub(crate) verbose: bool,
-    // Note: specifying multiple `--target` flags requires unstable `-Zmultitarget`,
+    // Note: specifying multiple `--target` flags requires unstable `-Z multitarget`,
     // so cargo-hack currently only supports a single `--target`.
     /// --target <TRIPLE>...
     pub(crate) target: Option<&'a str>,
 }
 
-pub(crate) fn raw() -> Vec<String> {
-    let mut args = env::args();
+pub(crate) fn raw() -> Result<Vec<String>> {
+    let mut args = env::args_os();
     let _ = args.next(); // executable name
-    args.collect()
+    handle_args(args)
 }
 
-pub(crate) fn parse_args<'a>(
-    raw: &'a [String],
-    cargo: &Cargo,
-    rustup: &Rustup,
-) -> Result<Args<'a>> {
+fn handle_args(args: impl IntoIterator<Item = impl Into<OsString>>) -> Result<Vec<String>> {
+    // Adapted from https://github.com/rust-lang/rust/blob/3bc9dd0dd293ab82945e35888ed6d7ab802761ef/compiler/rustc_driver/src/lib.rs#L1365-L1375.
+    args.into_iter()
+        .enumerate()
+        .map(|(i, arg)| {
+            arg.into()
+                .into_string()
+                .map_err(|arg| format_err!("argument {} is not valid Unicode: {:?}", i + 1, arg))
+        })
+        .collect()
+}
+
+pub(crate) fn parse_args<'a>(raw: &'a [String]) -> Result<(Args<'a>, Cargo)> {
     let mut iter = raw.iter();
     let args = &mut iter.by_ref().map(String::as_str).peekable();
     match args.next() {
@@ -293,13 +301,13 @@ pub(crate) fn parse_args<'a>(
             // TODO: implement
             warn!(
                 "--ignore-unknown-features for --include-features is not fully implemented and may not work as intended"
-            )
+            );
         }
         if !group_features.is_empty() {
             // TODO: implement
             warn!(
                 "--ignore-unknown-features for --group-features is not fully implemented and may not work as intended"
-            )
+            );
         }
     }
     if !each_feature && !feature_powerset {
@@ -428,13 +436,6 @@ pub(crate) fn parse_args<'a>(
         }
     }
 
-    if cargo.version < 41 && include_deps_features {
-        bail!("--include-deps-features requires Cargo 1.41 or leter");
-    }
-    if rustup.version < 23 && version_range.is_some() {
-        bail!("--version-range requires rustup 1.23 or leter");
-    }
-
     if subcommand.is_none() {
         if leading.contains(&"-h") {
             println!("{}", Help::short());
@@ -447,16 +448,26 @@ pub(crate) fn parse_args<'a>(
             print_version();
             std::process::exit(0);
         } else if leading.contains(&"--list") {
-            let mut line = cargo.process();
+            let mut line = Cargo::new().process();
             line.arg("--list");
-            line.exec()?;
+            line.run()?;
             std::process::exit(0);
         } else if !remove_dev_deps {
             // TODO: improve this
-            mini_usage("no subcommand or valid flag specified")?
+            mini_usage("no subcommand or valid flag specified")?;
         }
     }
 
+    let cargo = Cargo::new();
+    let rustup = Rustup::new();
+
+    if cargo.version < 41 && include_deps_features {
+        bail!("--include-deps-features requires Cargo 1.41 or later");
+    }
+    if rustup.version < 23 && version_range.is_some() {
+        bail!("--version-range requires rustup 1.23 or later");
+    }
+
     if no_dev_deps {
         info!(
             "--no-dev-deps removes dev-dependencies from real `Cargo.toml` while cargo-hack is running and restores it when finished"
@@ -467,42 +478,45 @@ pub(crate) fn parse_args<'a>(
     exclude_all_features |= !include_features.is_empty() || !exclude_features.is_empty();
     exclude_features.extend_from_slice(&features);
 
-    Ok(Args {
-        leading_args: leading,
-        trailing_args: iter.as_slice(),
+    Ok((
+        Args {
+            leading_args: leading,
+            trailing_args: iter.as_slice(),
 
-        subcommand,
+            subcommand,
 
-        manifest_path,
-        package,
-        exclude,
-        workspace: workspace.is_some(),
-        each_feature,
-        feature_powerset,
-        no_dev_deps,
-        remove_dev_deps,
-        ignore_private,
-        ignore_unknown_features,
-        optional_deps,
-        clean_per_run,
-        clean_per_version,
-        include_features: include_features.into_iter().map(Into::into).collect(),
-        include_deps_features,
-        version_range,
+            manifest_path,
+            package,
+            exclude,
+            workspace: workspace.is_some(),
+            each_feature,
+            feature_powerset,
+            no_dev_deps,
+            remove_dev_deps,
+            ignore_private,
+            ignore_unknown_features,
+            optional_deps,
+            clean_per_run,
+            clean_per_version,
+            include_features: include_features.into_iter().map(Into::into).collect(),
+            include_deps_features,
+            version_range,
 
-        depth,
-        group_features,
+            depth,
+            group_features,
 
-        exclude_features,
-        exclude_no_default_features,
-        exclude_all_features,
+            exclude_features,
+            exclude_no_default_features,
+            exclude_all_features,
 
-        features,
+            features,
 
-        no_default_features,
-        verbose,
-        target,
-    })
+            no_default_features,
+            verbose,
+            target,
+        },
+        cargo,
+    ))
 }
 
 fn parse_opt<'a>(
@@ -625,9 +639,13 @@ const HELP: &[HelpText<'_>] = &[
             "Note that ranges are always inclusive ranges.",
         ],
     ),
-    ("", "--version-step", "<NUM>", "Specify the version interval of --version-range", &[
-        "This flag can only be used together with --version-range flag.",
-    ]),
+    (
+        "",
+        "--version-step",
+        "<NUM>",
+        "Specify the version interval of --version-range (default to `1`)",
+        &["This flag can only be used together with --version-range flag."],
+    ),
     ("", "--clean-per-run", "", "Remove artifacts for that package before running the command", &[
         "If used this flag with --workspace, --each-feature, or --feature-powerset, artifacts will be removed before each run.",
         "Note that dependencies artifacts will be preserved.",
@@ -650,21 +668,15 @@ struct Help {
     print_version: bool,
 }
 
+const DEFAULT_TERM_SIZE: usize = 80;
+
 impl Help {
     fn long() -> Self {
-        Self {
-            long: true,
-            term_size: terminal_size::terminal_size().map_or(120, |(width, _)| width.0 as _),
-            print_version: true,
-        }
+        Self { long: true, term_size: DEFAULT_TERM_SIZE, print_version: true }
     }
 
     fn short() -> Self {
-        Self {
-            long: false,
-            term_size: terminal_size::terminal_size().map_or(120, |(width, _)| width.0 as _),
-            print_version: true,
-        }
+        Self { long: false, term_size: DEFAULT_TERM_SIZE, print_version: true }
     }
 }
 
@@ -754,7 +766,7 @@ Some common cargo commands are (see all commands with --list):
 }
 
 fn print_version() {
-    println!("{0} {1}", env!("CARGO_PKG_NAME"), env!("CARGO_PKG_VERSION"))
+    println!("{0} {1}", env!("CARGO_PKG_NAME"), env!("CARGO_PKG_VERSION"));
 }
 
 // Note: When adding a flag here, update the test with the same name in `tests/test.rs` file.
@@ -868,7 +880,7 @@ fn conflicts(a: &str, b: &str) -> Result<()> {
 
 #[cfg(test)]
 mod tests {
-    use std::{env, path::Path, process::Command};
+    use std::{env, panic, path::Path, process::Command};
 
     use anyhow::Result;
     use tempfile::Builder;
@@ -876,11 +888,28 @@ mod tests {
     use super::Help;
     use crate::fs;
 
+    // See handle_args function for more.
+    #[cfg(unix)]
+    #[test]
+    fn non_utf8_arg() {
+        use std::{ffi::OsStr, os::unix::prelude::OsStrExt};
+        // `cargo hack -- $'fo\x80o'`
+        super::handle_args(&[
+            "hack".as_ref(),
+            "--".as_ref(),
+            OsStr::from_bytes(&[b'f', b'o', 0x80, b'o']),
+        ])
+        .unwrap_err();
+    }
+
     #[track_caller]
     fn assert_diff(expected_path: impl AsRef<Path>, actual: impl AsRef<str>) {
         let actual = actual.as_ref();
         let manifest_dir = Path::new(env!("CARGO_MANIFEST_DIR"));
         let expected_path = &manifest_dir.join(expected_path);
+        if !expected_path.is_file() {
+            fs::write(expected_path, "").unwrap();
+        }
         let expected = fs::read_to_string(expected_path).unwrap();
         if expected != actual {
             if env::var_os("CI").is_some() {
@@ -922,7 +951,6 @@ mod tests {
         let mut start = false;
         let mut end = false;
         while let Some(line) = lines.next() {
-            dbg!(&line);
             out.push_str(line);
             out.push('\n');
             if line == "<!-- readme-long-help:start -->" {
@@ -930,7 +958,6 @@ mod tests {
                 out.push_str("```console\n");
                 out.push_str("$ cargo hack --help\n");
                 out.push_str(&new);
-                out.push('\n');
                 for line in &mut lines {
                     if line == "<!-- readme-long-help:end -->" {
                         out.push_str("```\n");
@@ -945,9 +972,9 @@ mod tests {
         if start && end {
             fs::write(path, out)?;
         } else if start {
-            panic!("do not modify `<!-- readme-long-help:end -->` comment in README.md")
+            panic!("missing `<!-- readme-long-help:end -->` comment in README.md");
         } else {
-            panic!("do not modify `<!-- readme-long-help:start -->` comment in README.md")
+            panic!("missing `<!-- readme-long-help:start -->` comment in README.md");
         }
         Ok(())
     }
